@@ -9,7 +9,8 @@ use std::str::FromStr;
 use async_trait::async_trait;
 use bytes::Bytes;
 use destream::de::Error as DestreamError;
-use destream::{Decoder, EncodeMap, Encoder, FromStream, IntoStream, ToStream};
+use destream::de::Visitor as DestreamVisitor;
+use destream::{Decoder, EncodeMap, Encoder, FromStream, IntoStream, MapAccess, ToStream};
 use log::debug;
 use safecast::{CastFrom, CastInto, TryCastFrom, TryCastInto};
 use serde::de::{Deserialize, Deserializer, Error as SerdeError};
@@ -280,21 +281,23 @@ impl TryCastFrom<Value> for ValueType {
     }
 }
 
+impl TryFrom<ValueType> for NumberType {
+    type Error = TCError;
+
+    fn try_from(vt: ValueType) -> TCResult<NumberType> {
+        match vt {
+            ValueType::Number(nt) => Ok(nt),
+            other => Err(TCError::bad_request("expected a Number type, not", other)),
+        }
+    }
+}
+
 #[async_trait]
 impl FromStream for ValueType {
     type Context = ();
 
-    async fn from_stream<D: Decoder>(cxt: (), decoder: &mut D) -> Result<Self, D::Error> {
-        let link = Link::from_stream(cxt, decoder).await?;
-        if let Some(host) = link.host() {
-            Err(DestreamError::invalid_value(
-                host,
-                "a Value type (without a host)",
-            ))
-        } else {
-            let path = link.into_path();
-            Self::from_path(&path).ok_or_else(|| DestreamError::invalid_value(path, "a Value type"))
-        }
+    async fn from_stream<D: Decoder>(_: (), decoder: &mut D) -> Result<Self, D::Error> {
+        decoder.decode_any(ValueTypeVisitor).await
     }
 }
 
@@ -313,6 +316,51 @@ impl fmt::Display for ValueType {
     }
 }
 
+struct ValueTypeVisitor;
+
+impl ValueTypeVisitor {
+    fn visit_path<E: DestreamError>(self, path: TCPathBuf) -> Result<ValueType, E> {
+        ValueType::from_path(&path)
+            .ok_or_else(|| DestreamError::invalid_value(path, Self::expecting()))
+    }
+}
+
+#[async_trait]
+impl DestreamVisitor for ValueTypeVisitor {
+    type Value = ValueType;
+
+    fn expecting() -> &'static str {
+        "a Value type"
+    }
+
+    fn visit_string<E: DestreamError>(self, v: String) -> Result<Self::Value, E> {
+        let path: TCPathBuf = v.parse().map_err(DestreamError::custom)?;
+        self.visit_path(path)
+    }
+
+    async fn visit_map<A: MapAccess>(self, mut access: A) -> Result<Self::Value, A::Error> {
+        if let Some(path) = access.next_key::<TCPathBuf>(()).await? {
+            if let Ok(list) = access.next_value::<Tuple<Value>>(()).await {
+                if list.is_empty() {
+                    self.visit_path(path)
+                } else {
+                    Err(DestreamError::invalid_value(list, "empty list"))
+                }
+            } else if let Ok(map) = access.next_value::<Map<Value>>(()).await {
+                if map.is_empty() {
+                    self.visit_path(path)
+                } else {
+                    Err(DestreamError::invalid_value(map, "empty map"))
+                }
+            } else {
+                Err(DestreamError::custom("invalid Value type"))
+            }
+        } else {
+            Err(DestreamError::invalid_length(0, Self::expecting()))
+        }
+    }
+}
+
 /// A generic value enum
 #[derive(Clone, Eq, PartialEq)]
 pub enum Value {
@@ -542,10 +590,10 @@ impl From<Number> for Value {
     }
 }
 
-impl From<Option<Value>> for Value {
-    fn from(opt: Option<Value>) -> Self {
+impl<T: Into<Value>> From<Option<T>> for Value {
+    fn from(opt: Option<T>) -> Self {
         match opt {
-            Some(value) => value,
+            Some(value) => value.into(),
             None => Self::None,
         }
     }
@@ -581,6 +629,23 @@ impl From<u64> for Value {
     }
 }
 
+impl<T1: CastInto<Value>, T2: CastInto<Value>> CastFrom<(T1, T2)> for Value {
+    fn cast_from(value: (T1, T2)) -> Self {
+        Value::Tuple(vec![value.0.cast_into(), value.1.cast_into()].into())
+    }
+}
+
+impl TryFrom<Value> for Number {
+    type Error = TCError;
+
+    fn try_from(value: Value) -> TCResult<Self> {
+        match value {
+            Value::Number(number) => Ok(number),
+            other => Err(TCError::bad_request("expected Number but found", other)),
+        }
+    }
+}
+
 impl TryFrom<Value> for Tuple<Value> {
     type Error = TCError;
 
@@ -646,7 +711,6 @@ impl TryCastFrom<Value> for Number {
             Value::Link(_) => false,
             Value::None => true,
             Value::Number(_) => true,
-            Value::Tuple(t) if t.is_empty() => true,
             Value::Tuple(t) if t.len() == 1 => Self::can_cast_from(&t[0]),
             Value::Tuple(t) if t.len() == 2 => {
                 Self::can_cast_from(&t[0]) && Self::can_cast_from(&t[1])
@@ -666,7 +730,6 @@ impl TryCastFrom<Value> for Number {
             Value::Link(_) => None,
             Value::None => Some(false.into()),
             Value::Number(n) => Some(n),
-            Value::Tuple(t) if t.is_empty() => Some(false.into()),
             Value::Tuple(mut t) if t.len() == 1 => Self::opt_cast_from(t.pop().unwrap()),
             Value::Tuple(mut t) if t.len() == 2 => {
                 let im = Self::opt_cast_from(t.pop().unwrap());
@@ -725,6 +788,16 @@ impl TryCastFrom<Value> for u64 {
     }
 }
 
+impl TryCastFrom<Value> for i64 {
+    fn can_cast_from(value: &Value) -> bool {
+        Number::can_cast_from(value)
+    }
+
+    fn opt_cast_from(value: Value) -> Option<Self> {
+        Number::opt_cast_from(value).map(|n| n.cast_into())
+    }
+}
+
 impl TryCastFrom<Value> for TCPathBuf {
     fn can_cast_from(value: &Value) -> bool {
         match value {
@@ -839,7 +912,23 @@ impl<T: Clone + TryCastFrom<Value>> TryCastFrom<Value> for Tuple<T> {
 
 impl fmt::Debug for Value {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        fmt::Display::fmt(self, f)
+        match self {
+            Self::Bytes(bytes) => write!(f, "({} bytes)", bytes.len()),
+            Self::Id(id) => write!(f, "{} {:?}", ValueType::Id, id.as_str()),
+            Self::Link(link) => write!(f, "{} {:?}", ValueType::Link, link),
+            Self::None => f.write_str("None"),
+            Self::Number(n) => fmt::Debug::fmt(n, f),
+            Self::String(s) => write!(f, "{}, {}", ValueType::String, s),
+            Self::Tuple(t) => write!(
+                f,
+                "{} ({})",
+                ValueType::Tuple,
+                t.iter()
+                    .map(|v| format!("{:?}", v))
+                    .collect::<Vec<String>>()
+                    .join(", ")
+            ),
+        }
     }
 }
 
